@@ -2,9 +2,11 @@
 # SPDX-License-Identifier: Apache-2.0
 
 import functools
+import importlib
 import inspect
 import math
 import typing
+import warnings
 from collections import OrderedDict
 from importlib import import_module
 
@@ -13,14 +15,14 @@ import makefun
 import funsor.delta
 import funsor.ops as ops
 from funsor.affine import is_affine
-from funsor.cnf import GaussianMixture
-from funsor.domains import Domain, reals
+from funsor.cnf import Contraction, GaussianMixture
+from funsor.domains import Array, Real, Reals
 from funsor.gaussian import Gaussian
 from funsor.interpreter import gensym
 from funsor.tensor import (Tensor, align_tensors, dummy_numeric_array, get_default_prototype,
                            ignore_jit_warnings, numeric_array, stack)
 from funsor.terms import Funsor, FunsorMeta, Independent, Number, Variable, eager, to_data, to_funsor
-from funsor.util import broadcast_shape, get_backend
+from funsor.util import broadcast_shape, get_backend, getargspec, lazy_property
 
 
 BACKEND_TO_DISTRIBUTIONS_BACKEND = {
@@ -31,7 +33,7 @@ BACKEND_TO_DISTRIBUTIONS_BACKEND = {
 
 def numbers_to_tensors(*args):
     """
-    Convert :class:`~funsor.terms.Number`s to :class:`funsor.tensor.Tensor`s,
+    Convert :class:`~funsor.terms.Number` s to :class:`funsor.tensor.Tensor` s,
     using any provided tensor as a prototype, if available.
     """
     if any(isinstance(x, Number) for x in args):
@@ -82,7 +84,7 @@ class Distribution(Funsor, metaclass=DistributionMeta):
             assert isinstance(value, Funsor)
             inputs.update(value.inputs)
         inputs = OrderedDict(inputs)
-        output = reals()
+        output = Real
         super(Distribution, self).__init__(inputs, output)
         self.params = OrderedDict(params)
 
@@ -103,6 +105,14 @@ class Distribution(Funsor, metaclass=DistributionMeta):
         data = cls.dist_class(**params).log_prob(value)
         return Tensor(data, inputs)
 
+    @property
+    def has_rsample(self):
+        return getattr(self.dist_class, "has_rsample", False)
+
+    @property
+    def has_enumerate_support(self):
+        return getattr(self.dist_class, "has_enumerate_support", False)
+
     def unscaled_sample(self, sampled_vars, sample_inputs, rng_key=None):
         params = OrderedDict(self.params)
         value = params.pop("value")
@@ -115,19 +125,31 @@ class Distribution(Funsor, metaclass=DistributionMeta):
 
         raw_dist = self.dist_class(**dict(zip(self._ast_fields[:-1], tensors)))
         sample_args = (sample_shape,) if rng_key is None else (rng_key, sample_shape)
-        if getattr(raw_dist, "has_rsample", False):
+        if self.has_rsample:
             raw_sample = raw_dist.rsample(*sample_args)
         else:
             raw_sample = ops.detach(raw_dist.sample(*sample_args))
 
         result = funsor.delta.Delta(value.name, Tensor(raw_sample, inputs, value.output.dtype))
-        if not getattr(raw_dist, "has_rsample", False):
+        if not self.has_rsample:
             # scaling of dice_factor by num samples should already be handled by Funsor.sample
             raw_log_prob = raw_dist.log_prob(raw_sample)
             dice_factor = Tensor(raw_log_prob - ops.detach(raw_log_prob), inputs)
             result = result + dice_factor
         return result
 
+    def enumerate_support(self, expand=False):
+        if not self.has_enumerate_support or not isinstance(self.value, Variable):
+            raise ValueError("cannot enumerate support of {}".format(repr(self)))
+        # arbitrary name-dim mapping, since we're converting back to a funsor anyway
+        name_to_dim = {name: -dim-1 for dim, (name, domain) in enumerate(self.inputs.items())
+                       if isinstance(domain.dtype, int) and name != self.value.name}
+        raw_dist = to_data(self, name_to_dim=name_to_dim)
+        raw_value = raw_dist.enumerate_support(expand=expand)
+        dim_to_name = {dim: name for name, dim in name_to_dim.items()}
+        dim_to_name[min(dim_to_name.keys(), default=0)-1] = self.value.name
+        return to_funsor(raw_value, output=self.value.output, dim_to_name=dim_to_name)
+
     def __getattribute__(self, attr):
         if attr in type(self)._ast_fields and attr != 'name':
             return self.params[attr]
@@ -144,7 +166,7 @@ class Distribution(Funsor, metaclass=DistributionMeta):
             out_dtype = int(instance.support.upper_bound + 1)
         else:
             out_dtype = 'real'
-        return Domain(dtype=out_dtype, shape=out_shape)
+        return Array[out_dtype, out_shape]
 
     @classmethod
     @functools.lru_cache(maxsize=5000)
@@ -156,17 +178,22 @@ class Distribution(Funsor, metaclass=DistributionMeta):
         # Because NumPyro and Pyro have the same pattern, we use name check for simplicity.
         support_name = type(support).__name__
         if support_name == "_Simplex":
-            output = reals(raw_shape[-1])
+            output = Reals[raw_shape[-1]]
         elif support_name == "_RealVector":
-            output = reals(raw_shape[-1])
+            output = Reals[raw_shape[-1]]
         elif support_name in ["_LowerCholesky", "_PositiveDefinite"]:
-            output = reals(*raw_shape[-2:])
+            output = Reals[raw_shape[-2:]]
         # resolve the issue: logits's constraints are real (instead of real_vector)
         # for discrete multivariate distributions in Pyro
-        elif support_name == "_Real" and name == "logits" and (
-                "probs" in cls.dist_class.arg_constraints
-                and type(cls.dist_class.arg_constraints["probs"]).__name__ == "_Simplex"):
-            output = reals(raw_shape[-1])
+        elif support_name == "_Real":
+            if name == "logits" and (
+                    "probs" in cls.dist_class.arg_constraints
+                    and type(cls.dist_class.arg_constraints["probs"]).__name__ == "_Simplex"):
+                output = Reals[raw_shape[-1]]
+            else:
+                output = Real
+        elif support_name in ("_Interval", "_GreaterThan", "_LessThan"):
+            output = Real
         else:
             output = None
         return output
@@ -205,7 +232,9 @@ FUNSOR_DIST_NAMES = [
     ('CategoricalLogits', ('logits',)),
     ('Delta', ('v', 'log_density')),
     ('Dirichlet', ('concentration',)),
+    ('DirichletMultinomial', ('concentration', 'total_count')),
     ('Gamma', ('concentration', 'rate')),
+    ('GammaPoisson', ('concentration', 'rate')),
     ('Multinomial', ('total_count', 'probs')),
     ('MultivariateNormal', ('loc', 'scale_tril')),
     ('NonreparameterizedBeta', ('concentration1', 'concentration0')),
@@ -213,7 +242,8 @@ FUNSOR_DIST_NAMES = [
     ('NonreparameterizedGamma', ('concentration', 'rate')),
     ('NonreparameterizedNormal', ('loc', 'scale')),
     ('Normal', ('loc', 'scale')),
-    ('Poisson', ('rate',))
+    ('Poisson', ('rate',)),
+    ('VonMises', ('loc', 'concentration')),
 ]
 
 
@@ -265,6 +295,75 @@ def mvndist_to_funsor(backend_dist, output=None, dim_to_name=None, real_inputs=O
     return discrete + Gaussian(gaussian.info_vec, gaussian.precision, inputs)
 
 
+class CoerceDistributionToFunsor:
+    """
+    Handler to reinterpret a backend distribution ``D`` as a corresponding
+    funsor during ``type(D).__call__()`` in case any constructor args are
+    funsors rather than backend tensors.
+
+    Example usage::
+
+        # in foo/distribution.py
+        coerce_to_funsor = CoerceDistributionToFunsor("foo")
+
+        class DistributionMeta(type):
+            def __call__(cls, *args, **kwargs):
+                result = coerce_to_funsor(cls, args, kwargs)
+                if result is not None:
+                    return result
+                return super().__call__(*args, **kwargs)
+
+        class Distribution(metaclass=DistributionMeta):
+            ...
+
+    :param str backend: Name of a funsor backend.
+    """
+    def __init__(self, backend):
+        self.backend = backend
+
+    @lazy_property
+    def module(self):
+        funsor.set_backend(self.backend)
+        module_name = BACKEND_TO_DISTRIBUTIONS_BACKEND[self.backend]
+        return importlib.import_module(module_name)
+
+    def __call__(self, cls, args, kwargs):
+        # Check whether distribution class takes any tensor inputs.
+        arg_constraints = getattr(cls, "arg_constraints", None)
+        if not arg_constraints:
+            return
+
+        # Check whether any tensor inputs are actually funsors.
+        try:
+            ast_fields = cls._funsor_ast_fields
+        except AttributeError:
+            ast_fields = cls._funsor_ast_fields = getargspec(cls.__init__)[0][1:]
+        kwargs = {name: value for pairs in (zip(ast_fields, args), kwargs.items())
+                  for name, value in pairs}
+        if not any(isinstance(value, (str, Funsor))
+                   for name, value in kwargs.items()
+                   if name in arg_constraints):
+            return
+
+        # Check for a corresponding funsor class.
+        try:
+            funsor_cls = cls._funsor_cls
+        except AttributeError:
+            funsor_cls = getattr(self.module, cls.__name__, None)
+            # resolve the issues Binomial/Multinomial are functions in NumPyro, which
+            # fallback to either BinomialProbs or BinomialLogits
+            if funsor_cls is None and cls.__name__.endswith("Probs"):
+                funsor_cls = getattr(self.module, cls.__name__[:-5], None)
+            cls._funsor_cls = funsor_cls
+        if funsor_cls is None:
+            warnings.warn("missing funsor for {}".format(cls.__name__),
+                          RuntimeWarning)
+            return
+
+        # Coerce to funsor.
+        return funsor_cls(**kwargs)
+
+
 ###############################################################
 # Converting distribution funsors to backend distributions
 ###############################################################
@@ -326,8 +425,10 @@ def Bernoulli(probs=None, logits=None, value='value'):
     """
     backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
     if probs is not None:
+        probs = to_funsor(probs, output=Real)
         return backend_dist.BernoulliProbs(probs, value)  # noqa: F821
     if logits is not None:
+        logits = to_funsor(logits, output=Real)
         return backend_dist.BernoulliLogits(logits, value)  # noqa: F821
     raise ValueError('Either probs or logits must be specified')
 
@@ -415,9 +516,9 @@ def eager_delta_variable_variable(v, log_density, value):
 
 
 def eager_normal(loc, scale, value):
-    assert loc.output == reals()
-    assert scale.output == reals()
-    assert value.output == reals()
+    assert loc.output == Real
+    assert scale.output == Real
+    assert value.output == Real
     if not is_affine(loc) or not is_affine(value):
         return None  # lazy
 
@@ -426,7 +527,7 @@ def eager_normal(loc, scale, value):
     log_prob = -0.5 * math.log(2 * math.pi) - ops.log(scale).sum()
     inputs = scale.inputs.copy()
     var = gensym('value')
-    inputs[var] = reals()
+    inputs[var] = Real
     gaussian = log_prob + Gaussian(info_vec, precision, inputs)
     return gaussian(**{var: value - loc})
 
@@ -444,6 +545,44 @@ def eager_mvn(loc, scale_tril, value):
     log_prob = -0.5 * scale_diag.shape[0] * math.log(2 * math.pi) - ops.log(scale_diag).sum()
     inputs = scale_tril.inputs.copy()
     var = gensym('value')
-    inputs[var] = reals(scale_diag.shape[0])
+    inputs[var] = Reals[scale_diag.shape[0]]
     gaussian = log_prob + Gaussian(info_vec, precision, inputs)
     return gaussian(**{var: value - loc})
+
+
+def eager_beta_bernoulli(red_op, bin_op, reduced_vars, x, y):
+    backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
+    return eager_dirichlet_multinomial(red_op, bin_op, reduced_vars, x,
+                                       backend_dist.Binomial(total_count=1, probs=y.probs, value=y.value))
+
+
+def eager_dirichlet_multinomial(red_op, bin_op, reduced_vars, x, y):
+    dirichlet_reduction = frozenset(x.inputs).intersection(reduced_vars)
+    if dirichlet_reduction:
+        backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
+        return backend_dist.DirichletMultinomial(concentration=x.concentration,
+                                                 total_count=y.total_count,
+                                                 value=y.value)
+    else:
+        return eager(Contraction, red_op, bin_op, reduced_vars, (x, y))
+
+
+def eager_gamma_poisson(red_op, bin_op, reduced_vars, x, y):
+    gamma_reduction = frozenset(x.inputs).intersection(reduced_vars)
+    if gamma_reduction:
+        backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
+        return backend_dist.GammaPoisson(concentration=x.concentration,
+                                         rate=x.rate,
+                                         value=y.value)
+    else:
+        return eager(Contraction, red_op, bin_op, reduced_vars, (x, y))
+
+
+def eager_dirichlet_posterior(op, c, z):
+    if (z.concentration is c.terms[0].concentration) and (c.terms[1].total_count is z.total_count):
+        backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
+        return backend_dist.Dirichlet(
+            concentration=z.concentration + c.terms[1].value,
+            value=c.terms[0].value)
+    else:
+        return None
